﻿using System;
using System.Collections.Generic;
using System.Linq;

namespace HeuristicLab.Problems.ProgramSynthesis.Push.Expressions {
  using System.Diagnostics;

  using Interpreter;

  public class PushProgram : StatefulExpression<Expression[]> {
    public static readonly PushProgram Empty = new PushProgram();

    private const string Prefix = "(";
    private const string PrefixReduced = "[";
    private const string Postfix = ")";
    private const string PostfixReduced = "]";
    private const string Delimiter = " ";

    public readonly bool IsEmpty;

    public PushProgram(params Expression[] state) : base(state) {
      IsEmpty = state.Length == 0;
    }

    public PushProgram(IEnumerable<Expression> state) : this(state.ToArray()) {
      // TODO: lazy evaluation of ToArray()
    }

    private string stringRepresentation;
    public override string StringRepresentation
    {
      get
      {
        if (string.IsNullOrEmpty(stringRepresentation))
          stringRepresentation = BuildString();
        return stringRepresentation;
      }
    }

    private int depth = -1;
    public int Depth
    {
      get
      {
        if (depth == -1) depth = CalcDepth();
        return depth;
      }
    }

    private int[] treeIndex;
    private int[] TreeIndex
    {
      get
      {
        return treeIndex ?? (treeIndex = BuildTreeIndex());
      }
    }

    private int hashCode;
    private int HashCode
    {
      get
      {
        if (hashCode == default(int)) hashCode = HashExpressions();
        return hashCode;
      }
    }

    //public PushProgram Copy() {
    //  // as the empty list is a static stateless expression, no copy required
    //  return ReferenceEquals(this, Empty) || this.State.IsEmpty
    //    ? Empty
    //    : new PushProgram(this.State.Copy());
    //}

    public override bool Eval(IPushInterpreter interpreter) {
      interpreter.ExecStack.Push(this.State);
      return true;
    }

    public override int GetHashCode() {
      return HashCode;
    }

    public override string ToString() {
      return this.StringRepresentation;
    }

    /// <summary>
    /// Returns the amount of elements in the whole tree
    /// </summary>
    /// <returns></returns>
    public int TotalCount
    {
      get
      {
        return this.IsEmpty
            ? 1
            // + 1 because "this" is also counted
            : TreeIndex[0] + 1;
      }
    }

    public IEnumerable<Expression> DepthFirst() {
      foreach (var expr in this.State) {
        if (expr.CanExpand) {
          var expandExpr = (PushProgram)expr;

          foreach (var sub in expandExpr.DepthFirst())
            yield return sub;
        } else yield return expr;
      }
    }

    public PushProgram Copy() {
      return IsEmpty ? this : new PushProgram(State) {
        stringRepresentation = this.stringRepresentation,
        hashCode = this.hashCode,
        depth = this.depth,
        treeIndex = this.treeIndex
      };
    }

    public Expression[] CopyExpressions()
    {
      if (IsEmpty) return Empty.State;

      var copy = new Expression[State.Length];
      State.CopyTo(copy, 0);

      return copy;
    }

    private int CalcDepth() {
      var maxDepth = 1;
      for (var i = 0; i < State.Length; i++) {
        var expression = State[i];
        if (!expression.CanExpand) continue;
        var expandExpression = (PushProgram)expression;

        if (expandExpression.Depth > maxDepth)
          maxDepth = expandExpression.Depth;
      }

      return maxDepth + 1;
    }

    private string BuildString() {
      // prevent too big string
      return this.TotalCount > 50
        ? PrefixReduced + this.State.Length + PostfixReduced
        : Prefix + string.Join(Delimiter, this.State.Where(e => !string.IsNullOrWhiteSpace(e.StringRepresentation))) + Postfix;
    }

    private int[] BuildTreeIndex() {
      var local_treeIndex = new int[State.Length];

      var next = 1;

      for (var i = State.Length - 1; i >= 0; i--) {
        var subExpression = State[i];

        local_treeIndex[i] = next;

        if (subExpression.CanExpand) {
          next += ((PushProgram)subExpression).TotalCount;
        } else {
          next++;
        }
      }

      return local_treeIndex;
    }

    private int HashExpressions() {
      var hash = 19 * 31 + this.GetType().FullName.GetHashCode();

      for (var i = 0; i < State.Length; i++) {
        hash = hash * 31 + State[i].GetHashCode();
      }

      return hash;
    }

    private static void CheckIfNested(int level, PushProgram target, PushProgram current,
        IList<PushProgram> visited) {
      if (visited.Any(x => ReferenceEquals(x, current)))
        Debugger.Break();

      visited.Add(current);
      if (level <= 0) {
        // hirarchy depth > level
        Debugger.Break();
        return;
      }

      for (var i = 0; i < current.State.Length; i++) {
        if (current.State[i].CanExpand)
          continue;

        var subExpression = current.State[i] as PushProgram;

        if (ReferenceEquals(target, subExpression)) {
          // Endless recursion dedected
          Debugger.Break();
        }

        CheckIfNested(level - 1, target, subExpression, visited);
      }

      visited.RemoveAt(visited.Count - 1);
    }

    /// <summary>
    ///     Returns the element with the given index whereby the tree is indexed in depth first manner
    /// </summary>
    /// <param name="index">The index of the required element</param>
    /// <returns>The required element</returns>
    public Expression GetFromTree(int index) {
      return GetFromTree(index, (root, parent, child, childIndex, parentIndex) => child);
    }

    /// <returns>The required element</returns>
    public T GetFromTree<T>(int index,
        Func<PushProgram, PushProgram, Expression, int, int, T> resolver) {
      return GetFromTree(index, 0, null, resolver);
    }

    private T GetFromTree<T>(int index, int parentIndex, PushProgram parent,
        Func<PushProgram, PushProgram, Expression, int, int, T> resolver) {
      if (index == 0) return resolver(parent, parent, this, 0, parentIndex);

      var min = State.Length - 1;
      var max = 0;
      var mid = (min + max) / 2;

      while ((max <= min) && (TreeIndex[mid] != index)) {
        if (TreeIndex[mid] > index) max = mid + 1;
        else min = mid - 1;

        mid = (min + max) / 2;
      }

      return TreeIndex[mid] == index
          ? resolver(parent, this, State[mid], mid, parentIndex)
          : (State[mid + 1] as PushProgram).GetFromTree(index - TreeIndex[mid + 1], mid + 1,
              this, resolver);
    }
  }
}
